/*
 * List related functions
 */
#include "s.h"
#include "vm2.h"
#include "stack.h"

static void wrong_arg_type(char *func, SOBJ obj)
{
  char erbuf[128];
  sprintf(erbuf, "%s: wrong arg type", func);
  err(erbuf, obj);
}

/* (pname pair?) */
SOBJ scm_pairp(SOBJ x)
{
  return (SCM_PAIRP(x) ? scm_true : scm_false);
}

SOBJ scm_cons(SOBJ car, SOBJ cdr)
{
  SOBJ new = scm_newcell(SOBJ_T_PAIR);
  SCM_CAR(new) = car;
  SCM_CDR(new) = cdr;
  return (new);
}

SOBJ scm_car(SOBJ obj)
{
  if (!SCM_PAIRP(obj))
    wrong_arg_type("car", obj);
  return (SCM_CAR(obj));
}

SOBJ scm_cdr(SOBJ obj)
{
  if (!SCM_PAIRP(obj))
    wrong_arg_type("cdr", obj);
  return (SCM_CDR(obj));
}

/* prim: (pname set-car!) */
SOBJ scm_setcar(SOBJ obj, SOBJ val)
{
  if (!SCM_PAIRP(obj))
    wrong_arg_type("set-car!", obj);
  SCM_CAR(obj) = val;
  return (scm_undefined);
}

/* (pname set-cdr!) */
SOBJ scm_setcdr(SOBJ obj, SOBJ val)
{
  if (!SCM_PAIRP(obj))
    wrong_arg_type("set-cdr!", obj);
  SCM_CDR(obj) = val;
  return (scm_undefined);
}

static SOBJ internal_cxr(SOBJ l, char *fct)
{
  register SOBJ tmp = l;
  register char *p;

  for (p = fct + strlen(fct) - 1; *p != 'X'; p--) {
    if (tmp == NULL || !SCM_PAIRP(tmp)) {
      char name[50];
      sprintf(name, "c%sr: bad list", fct + 1);
      err(name, l);
    }
    tmp = (*p == 'a') ? SCM_CAR(tmp) : SCM_CDR(tmp);
  }
  return tmp;
}

SOBJ scm_caar(SOBJ l) {  	return(internal_cxr(l, "Xaa"));		}
SOBJ scm_cdar(SOBJ l) {  	return(internal_cxr(l, "Xda"));		}
SOBJ scm_cadr(SOBJ l) {  	return(internal_cxr(l, "Xad"));		}
SOBJ scm_cddr(SOBJ l) {  	return(internal_cxr(l, "Xdd"));		}
SOBJ scm_caaar(SOBJ l) { 	return(internal_cxr(l, "Xaaa"));	}
SOBJ scm_cdaar(SOBJ l) {	return(internal_cxr(l, "Xdaa"));	}
SOBJ scm_cadar(SOBJ l) {	return(internal_cxr(l, "Xada"));	}
SOBJ scm_cddar(SOBJ l) {	return(internal_cxr(l, "Xdda"));	}
SOBJ scm_caadr(SOBJ l) {  	return(internal_cxr(l, "Xaad"));	}
SOBJ scm_cdadr(SOBJ l) { 	return(internal_cxr(l, "Xdad"));	}
SOBJ scm_caddr(SOBJ l) {  	return(internal_cxr(l, "Xadd"));	}
SOBJ scm_cdddr(SOBJ l) {  	return(internal_cxr(l, "Xddd"));	}
SOBJ scm_caaaar(SOBJ l) {  	return(internal_cxr(l, "Xaaaa"));	}
SOBJ scm_cdaaar(SOBJ l) {  	return(internal_cxr(l, "Xdaaa"));	}
SOBJ scm_cadaar(SOBJ l) {  	return(internal_cxr(l, "Xadaa"));	}
SOBJ scm_cddaar(SOBJ l) {  	return(internal_cxr(l, "Xddaa"));	}
SOBJ scm_caadar(SOBJ l) {  	return(internal_cxr(l, "Xaada"));	}
SOBJ scm_cdadar(SOBJ l) {   return(internal_cxr(l, "Xdada"));	}
SOBJ scm_caddar(SOBJ l) {   return(internal_cxr(l, "Xadda"));	}
SOBJ scm_cdddar(SOBJ l) {   return(internal_cxr(l, "Xddda"));	}
SOBJ scm_caaadr(SOBJ l) {   return(internal_cxr(l, "Xaaad"));	}
SOBJ scm_cdaadr(SOBJ l) {   return(internal_cxr(l, "Xdaad"));	}
SOBJ scm_cadadr(SOBJ l) {   return(internal_cxr(l, "Xadad"));	}
SOBJ scm_cddadr(SOBJ l) {   return(internal_cxr(l, "Xddad"));	}
SOBJ scm_caaddr(SOBJ l) {   return(internal_cxr(l, "Xaadd"));	}
SOBJ scm_cdaddr(SOBJ l) {   return(internal_cxr(l, "Xdadd"));	}
SOBJ scm_cadddr(SOBJ l) {   return(internal_cxr(l, "Xaddd"));	}
SOBJ scm_cddddr(SOBJ l) {   return(internal_cxr(l, "Xdddd"));	}

/* (pname null?) */
SOBJ scm_nullp(SOBJ obj)
{
  return ((obj == NULL) ? scm_true : scm_false);
}

/* Reset the gcmask bit. Note that only pair are supposed to have bit
 * set.
 */
static void scm_reset_list_mark(SOBJ list)
{

  while(list) {
	if (!SCM_CIRCBIT(list)) 	break;
	SCM_CIRCBIT_CLR(list);
	list = SCM_CDR(list);
  }
}

int scm_list_length(SOBJ list)
{
  SOBJ l;
  int len;

  /* first traversal: counting and setting the gcmark bit.  If the
   * gcmark bit is already set, then the list is circular and we stop
   * looping */

  for (len = 0, l = list; l != NULL; l = SCM_CDR(l), len++) {
	if (!SCM_PAIRP(l))	{ len = -1;  break; }
	if (SCM_CIRCBIT(l))		break; /* circular list */
	SCM_CIRCBIT_SET(l);
  }
  if (l != list) {
	scm_reset_list_mark(list);
  }
  return(len);
}

/* (pname list?) */
SOBJ scm_listp(SOBJ obj)
{
  return (scm_list_length(obj) >= 0 ? scm_true : scm_false);
}

SOBJ scm_list(int n, SOBJ *obja)
{
  SOBJ *p, list = NULL;
  p = obja + n;
  while (--p >= obja) {
    list = scm_cons(*p, list);
  }
  return (list);
}

SOBJ scm_length(SOBJ obj)
{
  int len;
  len = scm_list_length(obj);
  if (len < 0)
    err("length: not calculable", obj);
  return (SCM_MKINUM(len));
}

SOBJ scm_append2(SOBJ l1, SOBJ l2)
{
  SOBJ new, last, list;
  if (l1 == NULL)
    return (l2);
  if (!SCM_PAIRP(l1))
    err("append: argument is not a list", l1);

  list = scm_newcell(SOBJ_T_PAIR);
  new = list;
  last = NULL;
  while (1) {
    SCM_CAR(new) = SCM_CAR(l1);
    if (last) {
      SCM_CDR(last) = new;
    }
    last = new;
    l1 = SCM_CDR(l1);
    if (l1 == NULL || !SCM_PAIRP(l1))
      break;
    new = scm_newcell(SOBJ_T_PAIR);
  }
  SCM_CDR(new) = l2;
  return (list);
}

SOBJ scm_append(int len, SOBJ *l)
{
  switch (len) {
  case 0:
    return (NULL);
  case 1:
    return (l[0]);
  case 2:
    return scm_append2(l[0], l[1]);
  default:
    return scm_append2(l[0], scm_append(len - 1, l + 1));
  }
}

SOBJ scm_reverse(SOBJ l)
{
  SOBJ n = NULL;
  while (l) {
    if (!SCM_PAIRP(l))
      err("reverse: bad list", l);
    n = scm_cons(SCM_CAR(l), n);
    l = SCM_CDR(l);
  }
  return (n);
}

/* (pname list-tail) */
SOBJ scm_list_tail(SOBJ list, SOBJ k)
{
  SOBJ l;
  long x;

  if (!SCM_PAIRP(list))
    err("list-tail: bad list", list);
  if (!SCM_INUMP(k))
    err("list-tail: bad index", k);
  x = SCM_INUM(k);
  if (x < 0)
    err("list-tail: index must be exact positive integer", k);
  l = list;

  for (l = list; x > 0; x--) {
    if (l == NULL || !SCM_PAIRP(l))
      err("list-tail: list too short", list);
    l = SCM_CDR(l);
  }
  return l;
}

/* (pname list-ref) */
SOBJ scm_list_ref(SOBJ list, SOBJ k)
{
  SOBJ l;
  long x;

  if (!SCM_PAIRP(list))
    err("list-ref: Bad list", list);
  if (!SCM_INUMP(k))
    err("list-tail: bad index", k);
  x = SCM_INUM(k);
  if (x < 0)
    err("list-ref: index must be exact positive integer", k);
  l = list;

  for (; x > 0; x--) {
    if (l == NULL || !SCM_PAIRP(l))
      break;
    l = SCM_CDR(l);
  }
  if (l == NULL || !SCM_PAIRP(l))
    err("list-ref: list too short", list);

  return(SCM_CAR(l));
}

static SOBJ lmember(SOBJ obj, SOBJ list, SOBJ (*predicate) (SOBJ, SOBJ))
{
  register SOBJ ptr;

  if (!SCM_PAIRP(list) && !SCM_NULLP(list))
    err("member: bad list", list);
  for (ptr = list; !SCM_NULLP(ptr);) {
    if (SCM_PAIRP(ptr)) {
      if ((*predicate) (SCM_CAR(ptr), obj) == scm_true)
	return(ptr);
    } else
      /* end of a dotted list */
      return ((*predicate) (ptr, obj) == scm_true) ? ptr : scm_false;
    if ((ptr = SCM_CDR(ptr)) == list)
      err("member: circular list", NULL);
  }
  return(scm_false);
}

SOBJ scm_memq(SOBJ obj, SOBJ list)
{
  return(lmember(obj, list, scm_eq));
}

SOBJ scm_memv(SOBJ obj, SOBJ list)
{
  return(lmember(obj, list, scm_eqv));
}

SOBJ scm_member(SOBJ obj, SOBJ list)
{
  return(lmember(obj, list, scm_equal));
}


static SOBJ lassoc(SOBJ obj, SOBJ alist, SOBJ (*predicate) (SOBJ, SOBJ))
{
  register SOBJ l, tmp;

  for (l = alist; SCM_PAIRP(l);) {
    tmp = SCM_CAR(l);
    if (SCM_PAIRP(tmp) && (*predicate) (SCM_CAR(tmp), obj) == scm_true)
      return tmp;

    if ((l = SCM_CDR(l)) == alist)
      err("assoc function: cirular list", alist);
  }
  if (!SCM_NULLP(l))
    err("assoc function: improper list", alist);

  return (scm_false);
}

SOBJ scm_assq(SOBJ obj, SOBJ alist)
{
  return(lassoc(obj, alist, scm_eq));
}
SOBJ scm_assv(SOBJ obj, SOBJ alist)
{
  return(lassoc(obj, alist, scm_eqv));
}
SOBJ scm_assoc(SOBJ obj, SOBJ alist)
{
  return(lassoc(obj, alist, scm_equal));
}

SOBJ scm_map2(SOBJ func, int argc, SOBJ *argv, int map)
{
  int i;
  SOBJ res, *spsave, *tmp;

  res = NULL;
  tmp = &res;
  spsave = scm_sp;
  while(argv[0]) {					/* based on first list length */
	spush_opc(SCM_OP_END);
	spush_opc(SCM_OP_CALL);

	spush(func);
	for (i = 0; i < argc; i++) {
	  if (argv[i] == NULL) err("map: list too short", argv[i]);
	  spush(SCM_CAR(argv[i]));  argv[i] = SCM_CDR(argv[i]);
	}
	spush(SCM_MKINUM(argc + 1));	/* includes the func */
	spush_opc(SCM_OP_PUSHN);
	spush_opc(SCM_OP_MARK);
	if (map) {
	  *tmp = scm_cons(scm_run_engine(scm_sp), NULL);
	  tmp = &SCM_CDR(*tmp);
	} else {
	  scm_run_engine(scm_sp);
	}
	scm_sp = spsave;
  }
  return((map) ? res : scm_undefined);
}

SOBJ scm_map(int argc, SOBJ *argv)
{
  SOBJ func;
  if (argc < 2)	return(NULL);
  func = argv[0];
  scm_sp = argv;
  return(scm_map2(func, argc-1, argv+1, TRUE));
}

SOBJ scm_foreach(int argc, SOBJ *argv)
{
  SOBJ func;
  if (argc < 2)	return(NULL);
  func = argv[0];
  scm_sp = argv;
  return(scm_map2(func, argc-1, argv+1, FALSE));
}

void scm_init_list()
{
  scm_add_cprim("pair?"     , scm_pairp     , 1);
  scm_add_cprim("cons"      , scm_cons      , 2);
  scm_add_cprim("car"       , scm_car       , 1);
  scm_add_cprim("cdr"       , scm_cdr       , 1);
  scm_add_cprim("set-car!"  , scm_setcar    , 2);
  scm_add_cprim("set-cdr!"  , scm_setcdr    , 2);
  scm_add_cprim("caar"      , scm_caar      , 1);
  scm_add_cprim("cdar"      , scm_cdar      , 1);
  scm_add_cprim("cadr"      , scm_cadr      , 1);
  scm_add_cprim("cddr"      , scm_cddr      , 1);
  scm_add_cprim("caaar"     , scm_caaar     , 1);
  scm_add_cprim("cdaar"     , scm_cdaar     , 1);
  scm_add_cprim("cadar"     , scm_cadar     , 1);
  scm_add_cprim("cddar"     , scm_cddar     , 1);
  scm_add_cprim("caadr"     , scm_caadr     , 1);
  scm_add_cprim("cdadr"     , scm_cdadr     , 1);
  scm_add_cprim("caddr"     , scm_caddr     , 1);
  scm_add_cprim("cdddr"     , scm_cdddr     , 1);
  scm_add_cprim("caaaar"    , scm_caaaar    , 1);
  scm_add_cprim("cdaaar"    , scm_cdaaar    , 1);
  scm_add_cprim("cadaar"    , scm_cadaar    , 1);
  scm_add_cprim("cddaar"    , scm_cddaar    , 1);
  scm_add_cprim("caadar"    , scm_caadar    , 1);
  scm_add_cprim("cdadar"    , scm_cdadar    , 1);
  scm_add_cprim("caddar"    , scm_caddar    , 1);
  scm_add_cprim("cdddar"    , scm_cdddar    , 1);
  scm_add_cprim("caaadr"    , scm_caaadr    , 1);
  scm_add_cprim("cdaadr"    , scm_cdaadr    , 1);
  scm_add_cprim("cadadr"    , scm_cadadr    , 1);
  scm_add_cprim("cddadr"    , scm_cddadr    , 1);
  scm_add_cprim("caaddr"    , scm_caaddr    , 1);
  scm_add_cprim("cdaddr"    , scm_cdaddr    , 1);
  scm_add_cprim("cadddr"    , scm_cadddr    , 1);
  scm_add_cprim("cddddr"    , scm_cddddr    , 1);
  scm_add_cprim("null?"     , scm_nullp     , 1);
  scm_add_cprim("list?"     , scm_listp     , 1);
  scm_add_cprim("list"      , scm_list      , -1);
  scm_add_cprim("length"    , scm_length    , 1);
  scm_add_cprim("append2"   , scm_append2   , 2);
  scm_add_cprim("append"    , scm_append    , -1);
  scm_add_cprim("reverse"   , scm_reverse   , 1);
  scm_add_cprim("list-tail" , scm_list_tail , 2);
  scm_add_cprim("list-ref"  , scm_list_ref  , 2);
  scm_add_cprim("memq"      , scm_memq      , 2);
  scm_add_cprim("memv"      , scm_memv      , 2);
  scm_add_cprim("member"    , scm_member    , 2);
  scm_add_cprim("assq"      , scm_assq      , 2);
  scm_add_cprim("assv"      , scm_assv      , 2);
  scm_add_cprim("assoc"     , scm_assoc     , 2);
  scm_add_cprim("map"		, scm_map		, -1);
  scm_add_cprim("for-each"	, scm_foreach	, -1);
}
